#! /usr/bin/env python

import sys
import os
import csv
import collections

names_in_carstensen_but_not_in_jetz = [
    "Accipiter griseogularis",               # Accipitridae
    "Accipiter hiogaster",                   # Accipitridae
    "Acridotheres melanopterus",             # Sturnidae
    "Acrocephalus orientalis",               # Acrocephalidae
    "Aerodramus fuciphagus",                 # Apodidae
    "Aerodramus infuscatus",                 # Apodidae
    "Aerodramus vanikorensis",               # Apodidae
    "Anthus australis",                      # Motacillidae
    "Brachypteryx leucophris",               # Turdidae
    "Cacatua goffini",                       # Psittacidae
    "Cecropis striolata",                    # Hirundinidae
    "Chalcoparia singalensis",               # Nectariniidae
    "Cinnyris buettikoferi",                 # Nectariniidae
    "Cinnyris jugularis",                    # Nectariniidae
    "Cinnyris solaris",                      # Nectariniidae
    "Eos rubra",                             # Psittacidae
    "Euaegotheles crinifrons",               # Aegothelidae
    "Geopelia maugei",                       # Columbidae
    "Hieraaetus fasciatus",                  # Accipitridae
    "Hieraaetus kienerii",                   # Accipitridae
    "Leptocoma sericea",                     # Nectariniidae
    "Lichenostomus reticulatus",             # Meliphagidae
    "Lonchura fuscata",                      # Estrildidae
    "Megapodius forstenii",                  # Megapodiidae
    "Orthotomus cucullatus",                 # Cisticolidae
    "Pachycephala cinerea",                  # Pachysephalidae
    "Pachycephala fulvotincta",              # Pachysephalidae
    "Pachycephala macrorhyncha",             # Pachysephalidae
    "Pachycephala mentalis",                 # Pachysephalidae
    "Petrochelidon nigricans",               # Hirundinidae
    "Philemon plumigenis",                   # Meliphagidae
    "Ptilinopus nainus",                     # Columbidae
    "Reinwardtoena reinwardtii",             # Columbidae
    "Rhipidura dryas",                       # Rhipiduridae
    "Rhyticeros everetti",                   # Bucerotidae
    "Rhyticeros plicatus",                   # Bucerotidae
    "Species",                               # Family
    "Spizaetus cirrhatus",                   # Accipitridae
    "Spizaetus floris",                      # Accipitridae
    "Spizaetus lanceolatus",                 # Accipitridae
    "Streptopelia chinensis",                # Columbidae
    "Thapsinillas affinis",                  # Pycnonotidae
    "Thapsinillas longirostris",             # Pycnonotidae
    "Todiramphus saurophagus",               # Halcyonidae
    "Trichoglossus capistratus",             # Psittacidae
    "Trichoglossus forsteni",                # Psittacidae
    "Trichoglossus weberi",                  # Psittacidae
    "Zanclostomus calyorhynchus",            # Cuculidae
    "Zoothera horsfieldi",                   # Turdidae
]

carstensen_to_jetz_mapping = {
        "Acridotheres melanopterus": "Sturnus melanopterus",
        "Acrocephalus orientalis": "Acrocephalus arundinaceus",
        "Aerodramus fuciphagus": "Collocalia fuciphaga",
        "Aerodramus infuscatus": "Collocalia infuscata",
        "Aerodramus vanikorensis": "Collocalia vanikorensis",
        "Brachypteryx leucophris": "Brachypteryx leucophrys",
        "Cacatua goffini": "Cacatua goffiniana",
        "Cecropis striolata": "Hirundo striolata",
        "Chalcoparia singalensis": "Anthreptes singalensis",
        "Cinnyris buettikoferi": "Nectarinia buettikoferi",
        "Cinnyris jugularis": "Nectarinia jugularis",
        "Cinnyris solaris": "Nectarinia solaris",
        "Eos rubra": "Eos bornea",
        "Euaegotheles crinifrons": "Aegotheles crinifrons",
        "Geopelia maugei": "Geopelia maugeus",
        "Hieraaetus fasciatus": "Aquila fasciatus",
        "Hieraaetus kienerii": "Lophotriorchis kienerii",
        "Leptocoma sericea": "Nectarinia aspasia",
        "Lonchura fuscata": "Padda fuscata",
        "Orthotomus cucullatus": "Orthotomus cuculatus",
        "Petrochelidon nigricans": "Hirundo nigricans",
        "Ptilinopus nainus": "Ptilinopus naina",
        "Reinwardtoena reinwardtii": "Reinwardtoena reinwardtsi",
        "Rhyticeros everetti": "Aceros everetti",
        "Rhyticeros plicatus": "Aceros plicatus",
        "Spizaetus cirrhatus": "Nisaetus cirrhatus",
        "Spizaetus floris": "Nisaetus floris",
        "Spizaetus lanceolatus": "Nisaetus lanceolatus",
        "Streptopelia chinensis": "Stigmatopelia chinensis",
        "Thapsinillas affinis": "Alophoixus affinis",
        "Todiramphus saurophagus": "Todiramphus saurophaga",
        "Zanclostomus calyorhynchus": "Phaenicophaeus calyorhynchus",
        "Zoothera horsfieldi": "Myophonus horsfieldii",
        }

carstensen_to_jetz_mapping_uncertain = {
        # "Pachycephala cinerea": ("Pachycephala albiventris", "Pachycephala homeyeri", "Pachycephala grisola",),
        "Pachycephala cinerea": "Pachycephala grisola",
        "Pachycephala fulvotincta": ("Pachycephala pectoralis", "Pachycephala melanura", "Pachycephala jacquinoti",),
        "Pachycephala macrorhyncha": ("Pachycephala pectoralis", "Pachycephala melanura", "Pachycephala jacquinoti",),
        "Pachycephala mentalis": ("Pachycephala pectoralis", "Pachycephala melanura", "Pachycephala jacquinoti",),
        "Trichoglossus capistratus": ("Trichoglossus haematodus", "Trichoglossus rosenbergii", "Trichoglossus moluccanus", "Trichoglossus forsteni", "Trichoglossus capistratus", "Trichoglossus weberii", "Trichoglossus rubritorquis",),
        "Trichoglossus forsteni": ("Trichoglossus haematodus", "Trichoglossus rosenbergii", "Trichoglossus moluccanus", "Trichoglossus forsteni", "Trichoglossus capistratus", "Trichoglossus weberii", "Trichoglossus rubritorquis",),
        "Trichoglossus weberi": ("Trichoglossus haematodus", "Trichoglossus rosenbergii", "Trichoglossus moluccanus", "Trichoglossus forsteni", "Trichoglossus capistratus", "Trichoglossus weberii", "Trichoglossus rubritorquis",),
        }

carstensen_to_jetz_mapping_split_out_and_overrepresented_species = {
        "Anthus australis": "Anthus novaeseelandiae",
        "Philemon plumigenis": "Philemon moluccensis", # NOTE: 'Philemon moluccensis' already has a separate entry in original Carstensen database"
        "Rhipidura dryas": "Rhipidura rufifrons",
        "Megapodius forstenii": "Megapodius reinwardt",
        "Thapsinillas longirostris": "Alphoixus affinis",
        "Accipiter hiogaster": ("Accipiter hiogaster","Accipiter griseogularis", "Accipiter fasciatus",),
        "Accipiter griseogularis": ("Accipiter hiogaster","Accipiter griseogularis", "Accipiter fasciatus",),
        "Lichenostomus reticulatus": "Meliphaga reticulata",
        }

module_names = [
    "module_area_Banda_Sea",
    "module_area_Lesser_Sundas",
    "module_area_Maluku",
    "module_area_Sulawesi",
]

island_to_module_map = {
    "Adonara"                            : "Lesser_Sundas",
    "Alor"                               : "Lesser_Sundas",
    "Ambelau"                            : "Banda_Sea",
    "Ambon"                              : "Maluku",
    "Autauro"                            : "Lesser_Sundas", # Atauro, Makili
    "Bacan"                              : "Maluku",
    "Banda_islands"                      : "Banda_Sea",
    "Banggai_islands"                    : "Sulawesi",
    "Barbar"                             : "Banda_Sea",
    "Bisa"                               : "Maluku", # Obi islands
    "Boano"                              : "Maluku",
    "Bonerate"                           : "Banda_Sea",
    "Buru"                               : "Maluku",
    "Butung_or_Buton"                    : "Sulawesi",
    "Damar_islands"                      : "Banda_Sea", # In Maluku province
    "Flores"                             : "Lesser_Sundas",
    "Gomumu_island"                      : "Maluku", # Obi islands
    "Gorong_islands"                     : "Maluku",
    "Halmahera"                          : "Maluku",
    "Haruku"                             : "Maluku",
    "Kabaena"                            : "Sulawesi",
    "Kai_Islands"                        : "Banda_Sea",
    "Kalaotoa_islands"                   : "Banda_Sea",
    "Kasiruta"                           : "Sulawesi",
    "Kayoa"                              : "Maluku",
    "Kisar"                              : "Banda_Sea",
    "Kofiau"                             : "Maluku",
    "Komodo"                             : "Lesser_Sundas",
    "Lembeh"                             : "Sulawesi",
    "Leti_islands"                       : "Banda_Sea",
    "Loloda_islands"                     : "Maluku",
    "Lomblen"                            : "Lesser_Sundas", # Lembata
    "Lombok"                             : "Lesser_Sundas",
    "Lucipara_islands"                   : "Banda_Sea",
    "Madu"                               : "Banda_Sea",
    "Makian"                             : "Maluku",
    "Manadotua"                          : "Sulawesi",
    "Manipa"                             : "Maluku",
    "Manuk_island"                       : "Banda_Sea",
    "Mare"                               : "Maluku",
    "Mayu_island"                        : "Maluku",
    "Miangas"                            : "Sulawesi", # Talaud
    "Moa"                                : "Banda_Sea",
    "Morotai"                            : "Maluku",
    "Moti"                               : "Maluku",
    "Muna"                               : "Sulawesi",
    "Muor"                               : "Maluku",
    "Nila"                               : "Banda_Sea",
    "Nusa_Laut"                          : "Maluku",
    "Obi_islands"                        : "Maluku",
    "Padar"                              : "Lesser_Sundas",
    "Palue_island"                       : "Lesser_Sundas", # Palu
    "Pantar"                             : "Lesser_Sundas",
    "Rau_island"                         : "Lesser_Sundas",
    "Rinca"                              : "Lesser_Sundas",
    "Romang_islands"                     : "Banda_Sea",
    "Roti"                               : "Lesser_Sundas",
    "Sangeang"                           : "Lesser_Sundas",
    "Sangihe_Islands"                    : "Sulawesi",
    "Saparua"                            : "Maluku",
    "Satonda"                            : "Lesser_Sundas",
    "Sawu"                               : "Lesser_Sundas",
    "Selayar"                            : "Banda_Sea",
    "Semau"                              : "Lesser_Sundas",
    "Seram"                              : "Maluku",
    "Sermata_islands"                    : "Banda_Sea",
    "Siau"                               : "Sulawesi",
    "Solor"                              : "Lesser_Sundas",
    "Sula_islands"                       : "Sulawesi",
    "Sulawesi"                           : "Sulawesi",
    "Sumba"                              : "Lesser_Sundas",
    "Sumbawa"                            : "Lesser_Sundas",
    "Tobea_island"                       : "Sulawesi",
    "Togian_islands"                     : "Sulawesi",
    "Tukangbesi_islands"                 : "Banda_Sea",
    "Talaud_islands"                     : "Sulawesi",
    "Taliabu_westernPartOf_Sula_Islands" : "Sulawesi",
    "Talisei_island"                     : "Sulawesi",
    "Tanahjampea"                        : "Banda_Sea",
    "Tanimbar_Islands"                   : "Banda_Sea",
    "Tayandu_islands"                    : "Banda_Sea",
    "Ternate"                            : "Maluku",
    "Teun"                               : "Banda_Sea",
    "Tidore"                             : "Maluku",
    "Tifore"                             : "Maluku",
    "Timor_island_whole"                 : "Lesser_Sundas",
    "Tioor_or_Tiur"                      : "Banda_Sea",
    "Watubela_Islands"                   : "Banda_Sea",
    "Wetar"                              : "Lesser_Sundas",
    "Widi_islands"                       : "Maluku",
}

src_to_master_field_map = {
    "Number"                                   : None,
    "Family"                                   : "family",
    "Species"                                  : "carstensen_name",
    "Vernacular"                               : None,
    "Adonara.Latin"                            : "island_area_Adonara",
    "Alor.Latin"                               : "island_area_Alor",
    "Ambelau.Latin"                            : "island_area_Ambelau",
    "Ambon.Latin"                              : "island_area_Ambon",
    "Autauro.Latin"                            : "island_area_Autauro",
    "Bacan.Latin"                              : "island_area_Bacan",
    "Banda_islands.Latin"                      : "island_area_Banda_islands",
    "Banggai_islands.Latin"                    : "island_area_Banggai_islands",
    "Barbar.Latin"                             : "island_area_Barbar",
    "Bisa.Latin"                               : "island_area_Bisa",
    "Boano.Latin"                              : "island_area_Boano",
    "Bonerate.Latin"                           : "island_area_Bonerate",
    "Buru.Latin"                               : "island_area_Buru",
    "Butung_or_Buton.Latin"                    : "island_area_Butung_or_Buton",
    "Damar_islands.Latin"                      : "island_area_Damar_islands",
    "Flores.Latin"                             : "island_area_Flores",
    "Gomumu_island.Latin"                      : "island_area_Gomumu_island",
    "Gorong_islands.Latin"                     : "island_area_Gorong_islands",
    "Halmahera.Latin"                          : "island_area_Halmahera",
    "Haruku.Latin"                             : "island_area_Haruku",
    "Kabaena.Latin"                            : "island_area_Kabaena",
    "Kai_Islands.Latin"                        : "island_area_Kai_Islands",
    "Kalaotoa_islands.Latin"                   : "island_area_Kalaotoa_islands",
    "Kasiruta.Latin"                           : "island_area_Kasiruta",
    "Kayoa.Latin"                              : "island_area_Kayoa",
    "Kisar.Latin"                              : "island_area_Kisar",
    "Kofiau.Latin"                             : "island_area_Kofiau",
    "Komodo.Latin"                             : "island_area_Komodo",
    "Lembeh.Latin"                             : "island_area_Lembeh",
    "Leti_islands.Latin"                       : "island_area_Leti_islands",
    "Loloda_islands.Latin"                     : "island_area_Loloda_islands",
    "Lomblen.Latin"                            : "island_area_Lomblen",
    "Lombok.Latin"                             : "island_area_Lombok",
    "Lucipara_islands.Latin"                   : "island_area_Lucipara_islands",
    "Madu.Latin"                               : "island_area_Madu",
    "Makian.Latin"                             : "island_area_Makian",
    "Manadotua.Latin"                          : "island_area_Manadotua",
    "Manipa.Latin"                             : "island_area_Manipa",
    "Manuk_island.Latin"                       : "island_area_Manuk_island",
    "Mare.Latin"                               : "island_area_Mare",
    "Mayu_island.Latin"                        : "island_area_Mayu_island",
    "Miangas.Latin"                            : "island_area_Miangas",
    "Moa.Latin"                                : "island_area_Moa",
    "Morotai.Latin"                            : "island_area_Morotai",
    "Moti.Latin"                               : "island_area_Moti",
    "Muna.Latin"                               : "island_area_Muna",
    "Muor.Latin"                               : "island_area_Muor",
    "Nila.Latin"                               : "island_area_Nila",
    "Nusa_Laut.Latin"                          : "island_area_Nusa_Laut",
    "Obi_islands.Latin"                        : "island_area_Obi_islands",
    "Padar.Latin"                              : "island_area_Padar",
    "Palu'e_island.Latin"                      : "island_area_Palue_island",
    "Pantar.Latin"                             : "island_area_Pantar",
    "Rau_island.Latin"                         : "island_area_Rau_island",
    "Rinca.Latin"                              : "island_area_Rinca",
    "Romang_islands.Latin"                     : "island_area_Romang_islands",
    "Roti.Latin"                               : "island_area_Roti",
    "Sangeang.Latin"                           : "island_area_Sangeang",
    "Sangihe_Islands.Latin"                    : "island_area_Sangihe_Islands",
    "Saparua.Latin"                            : "island_area_Saparua",
    "Satonda.Latin"                            : "island_area_Satonda",
    "Sawu.Latin"                               : "island_area_Sawu",
    "Selayar.Latin"                            : "island_area_Selayar",
    "Semau.Latin"                              : "island_area_Semau",
    "Seram.Latin"                              : "island_area_Seram",
    "Sermata_islands.Latin"                    : "island_area_Sermata_islands",
    "Siau.Latin"                               : "island_area_Siau",
    "Solor.Latin"                              : "island_area_Solor",
    "Sula_islands.Latin"                       : "island_area_Sula_islands",
    "Sulawesi.Latin"                           : "island_area_Sulawesi",
    "Sumba.Latin"                              : "island_area_Sumba",
    "Sumbawa.Latin"                            : "island_area_Sumbawa",
    "Tobea_island.Latin"                       : "island_area_Tobea_island",
    "Togian_islands.Latin"                     : "island_area_Togian_islands",
    "Tukangbesi_islands.Latin"                 : "island_area_Tukangbesi_islands",
    "Talaud_islands.Latin"                     : "island_area_Talaud_islands",
    "Taliabu_westernPartOf_Sula_Islands.Latin" : "island_area_Taliabu_westernPartOf_Sula_Islands",
    "Talisei_island.Latin"                     : "island_area_Talisei_island",
    "Tanahjampea.Latin"                        : "island_area_Tanahjampea",
    "Tanimbar_Islands.Latin"                   : "island_area_Tanimbar_Islands",
    "Tayandu_islands.Latin"                    : "island_area_Tayandu_islands",
    "Ternate.Latin"                            : "island_area_Ternate",
    "Teun.Latin"                               : "island_area_Teun",
    "Tidore.Latin"                             : "island_area_Tidore",
    "Tifore.Latin"                             : "island_area_Tifore",
    "Timor_island_whole.Latin"                 : "island_area_Timor_island_whole",
    "Tioor_or_Tiur.Latin"                      : "island_area_Tioor_or_Tiur",
    "Watubela_Islands.Latin"                   : "island_area_Watubela_Islands",
    "Wetar.Latin"                              : "island_area_Wetar",
    "Widi_islands.Latin"                       : "island_area_Widi_islands",
    "Node"                                     : "node",
    "Familiy"                                  : None,
    "Links"                                    : "links",
    "r"                                        : "r",
    "l"                                        : "l",
    "Module"                                   : "module",
    "Role"                                     : "role",
    "Guild"                                    : "guild",
    "IUCN status"                              : None,
    "Mean Lenght"                              : None,
    "Mean Mass"                                : None,
    "Min altitude"                             : None,
    "Max altitude"                             : None,
    "Altitude range"                           : None,
    "Habitat category"                         : "habitat",
    "Reference"                                : None,
}

master_fieldnames = [
    "study_status",
    "carstensen_name",
    "jetz_name",
    "operational_name",
    "family",
    "habitat",
    "role",
    "guild",
    "module",
    "module_area_Banda_Sea",
    "module_area_Lesser_Sundas",
    "module_area_Maluku",
    "module_area_Sulawesi",
    "island_area_Adonara",
    "island_area_Alor",
    "island_area_Ambelau",
    "island_area_Ambon",
    "island_area_Autauro",
    "island_area_Bacan",
    "island_area_Banda_islands",
    "island_area_Banggai_islands",
    "island_area_Barbar",
    "island_area_Bisa",
    "island_area_Boano",
    "island_area_Bonerate",
    "island_area_Buru",
    "island_area_Butung_or_Buton",
    "island_area_Damar_islands",
    "island_area_Flores",
    "island_area_Gomumu_island",
    "island_area_Gorong_islands",
    "island_area_Halmahera",
    "island_area_Haruku",
    "island_area_Kabaena",
    "island_area_Kai_Islands",
    "island_area_Kalaotoa_islands",
    "island_area_Kasiruta",
    "island_area_Kayoa",
    "island_area_Kisar",
    "island_area_Kofiau",
    "island_area_Komodo",
    "island_area_Lembeh",
    "island_area_Leti_islands",
    "island_area_Loloda_islands",
    "island_area_Lomblen",
    "island_area_Lombok",
    "island_area_Lucipara_islands",
    "island_area_Madu",
    "island_area_Makian",
    "island_area_Manadotua",
    "island_area_Manipa",
    "island_area_Manuk_island",
    "island_area_Mare",
    "island_area_Mayu_island",
    "island_area_Miangas",
    "island_area_Moa",
    "island_area_Morotai",
    "island_area_Moti",
    "island_area_Muna",
    "island_area_Muor",
    "island_area_Nila",
    "island_area_Nusa_Laut",
    "island_area_Obi_islands",
    "island_area_Padar",
    "island_area_Palue_island",
    "island_area_Pantar",
    "island_area_Rau_island",
    "island_area_Rinca",
    "island_area_Romang_islands",
    "island_area_Roti",
    "island_area_Sangeang",
    "island_area_Sangihe_Islands",
    "island_area_Saparua",
    "island_area_Satonda",
    "island_area_Sawu",
    "island_area_Selayar",
    "island_area_Semau",
    "island_area_Seram",
    "island_area_Sermata_islands",
    "island_area_Siau",
    "island_area_Solor",
    "island_area_Sula_islands",
    "island_area_Sulawesi",
    "island_area_Sumba",
    "island_area_Sumbawa",
    "island_area_Tobea_island",
    "island_area_Togian_islands",
    "island_area_Tukangbesi_islands",
    "island_area_Talaud_islands",
    "island_area_Taliabu_westernPartOf_Sula_Islands",
    "island_area_Talisei_island",
    "island_area_Tanahjampea",
    "island_area_Tanimbar_Islands",
    "island_area_Tayandu_islands",
    "island_area_Ternate",
    "island_area_Teun",
    "island_area_Tidore",
    "island_area_Tifore",
    "island_area_Timor_island_whole",
    "island_area_Tioor_or_Tiur",
    "island_area_Watubela_Islands",
    "island_area_Wetar",
    "island_area_Widi_islands",
    "node",
    "links",
    "r",
    "l",
]

jetz_taxonomy_src_filepath = "jetz/taxonomy/jetz_taxonomy_species_names.txt"
carstensen_src_filepath = "carstensen/carstensen.csv"
jetz_trees_filepath = "jetz/trees/jetz_hackett1.tre"
master_reference_filepath = "master_reference.csv"

def open_source_file(filepath):
    if sys.version_info >= (3,0,0):
        f = open(filepath, 'r')
    else:
        f = open(filepath, 'rU')
    return f

def open_output(filepath):
    if filepath is None:
        return sys.stdout
    f = open(filepath, "w")
    return f

def open_output_file_for_csv_writer(filepath, append=False):
    if filepath is None:
        out = sys.stdout
    elif sys.version_info >= (3,0,0):
        out = open(filepath,
                "a" if append else "w",
                newline='')
    else:
        out = open(filepath, "ab" if append else "wb")
    return out

def build_master_reference():
    jetz_names = set(open(jetz_taxonomy_src_filepath, "r").read().split("\n"))
    jetz_to_operational_name_map = {}
    src = csv.DictReader(open_source_file(carstensen_src_filepath))
    out_rows = []
    master_island_fieldnames = [f for f in master_fieldnames if f.startswith("island_area")]
    master_module_fieldnames = [f for f in master_fieldnames if f.startswith("module_area")]
    module_taxa = collections.Counter()
    num_modules_lineage_occurs = collections.Counter()
    num_included_species = 0
    num_excluded_species = 0
    for row_idx, src_row in enumerate(src):
        out_row = {}
        for src_key in src_row:
            out_key = src_to_master_field_map[src_key]
            if out_key is None:
                continue
            out_row[out_key] = src_row[src_key]
        src_name = out_row["carstensen_name"]
        out_row["study_status"] = "included"
        out_row["operational_name"] = src_name.replace(" ", ".").replace("_", ".")
        if src_name in jetz_names:
            out_row["jetz_name"] = src_name
        elif src_name in carstensen_to_jetz_mapping:
            out_row["jetz_name"] = carstensen_to_jetz_mapping[src_name]
        elif ((src_name in carstensen_to_jetz_mapping_uncertain)
                or (src_name in carstensen_to_jetz_mapping_split_out_and_overrepresented_species)
            ):
            out_row["jetz_name"] = "NA"
            out_row["study_status"] = "excluded (name mapping failure)"
        else:
            raise ValueError(src_name)

        if not out_row["habitat"]:
            sys.stderr.write("Excluding '{}': no habitat information\n".format(src_name))
            out_row["study_status"] = "excluded (no habitat information)"
        if out_row["habitat"] == "6":
            sys.stderr.write("Excluding '{}': non-informative habitat category\n".format(src_name))
            out_row["study_status"] = "excluded (non-informative habitat category)"
        elif not out_row["guild"]:
            sys.stderr.write("Excluding '{}': no guild information\n".format(src_name))
            out_row["study_status"] = "excluded (no guild information)"
        elif not out_row["role"]:
            sys.stderr.write("Excluding '{}': no role information\n".format(src_name))
            out_row["study_status"] = "excluded (no role information)"
        if out_row["study_status"] == "included":
            jetz_to_operational_name_map[out_row["jetz_name"]] = out_row["operational_name"]

        for module_fieldname in master_module_fieldnames:
            out_row[module_fieldname] = "0"
        for island_fieldname in master_island_fieldnames:
            if out_row[island_fieldname] == "1":
                key = island_fieldname.replace("island_area_", "")
                module_name = island_to_module_map[key]
                module_key = "module_area_" + module_name
                assert module_key in out_row
                out_row[module_key] = "1"
            else:
                assert out_row[island_fieldname] == "0", out_row[island_fieldname]
        num_module_occurrences = 0
        for module_fieldname in master_module_fieldnames:
            if out_row[module_fieldname] == "1":
                num_module_occurrences += 1
                if out_row["study_status"] == "included":
                    module_taxa[module_fieldname] += 1
        assert num_module_occurrences > 0
        if out_row["study_status"] == "included":
            num_modules_lineage_occurs[num_module_occurrences] += 1
            num_included_species += 1
        else:
            num_excluded_species += 1
        out_rows.append(out_row)
    out = open_output_file_for_csv_writer(master_reference_filepath)
    writer = csv.DictWriter(
            out,
            fieldnames=master_fieldnames,
            delimiter=",",
            lineterminator=os.linesep,
    )
    writer.writeheader()
    writer.writerows(out_rows)

    print("--")
    print("Number of included species: {}".format(num_included_species))
    print("Number of excluded species: {}".format(num_excluded_species))
    print("Module Diversity: {}".format(module_taxa))
    print("Number of Modules Species Occur in: {}".format(num_modules_lineage_occurs))

    return jetz_to_operational_name_map

def build():
    jetz_to_operational_name_map = build_master_reference()

if __name__ == "__main__":
    build()
